# -*- coding: utf-8 -*-
"""
Created on Sun Aug 22 13:34:42 2020

@author: pankaj

------------------------------------------------------------------------

"""
import pytest

from radis import SpectrumFactory, get_residual
from radis.misc.printer import printm
from radis.test.utils import getTestFile


def test_eq_spectrum_emulated_gpu(emulate=True, plot=False, *args, **kwargs):

    print("Emulate: ", emulate)
    T = 1000
    p = 0.1
    wstep = 0.001
    wmin = 2284.0  # cm-1
    wmax = 2285.0  # cm-1

    sf = SpectrumFactory(
        wavenum_min=wmin,
        wavenum_max=wmax,
        mole_fraction=0.01,  # until self and air broadening is implemented
        path_length=1,  # doesnt change anything
        wstep=wstep,
        pressure=p,
        isotope="1,2",
        warnings={
            "MissingSelfBroadeningWarning": "ignore",
            "NegativeEnergiesWarning": "ignore",
            "HighTemperatureWarning": "ignore",
            "GaussianBroadeningWarning": "ignore",
        },
    )
    sf._broadening_method = "fft"
    sf.load_databank(
        path=getTestFile("cdsd_hitemp_09_fragment.txt"),
        format="cdsd-4000",
        parfuncfmt="hapi",
    )

    s_cpu = sf.eq_spectrum(Tgas=T, name="CPU")
    s_gpu = sf.eq_spectrum_gpu(
        Tgas=T, emulate=emulate, name="GPU (emulate)" if emulate else "GPU"
    )
    s_cpu.crop(wmin=2284.2, wmax=2284.8)  # remove edge lines
    s_gpu.crop(wmin=2284.2, wmax=2284.8)
    if plot:
        s_cpu.compare_with(s_gpu, spectra_only=True, plot=plot)
    assert get_residual(s_cpu, s_gpu, "abscoeff") < 1.4e-5
    assert get_residual(s_cpu, s_gpu, "radiance_noslit") < 7.3e-6
    assert get_residual(s_cpu, s_gpu, "transmittance_noslit") < 1.4e-5


@pytest.mark.needs_cuda
def test_eq_spectrum_gpu(plot, *args, **kwargs):
    test_eq_spectrum_emulated_gpu(emulate=False, plot=plot, *args, **kwargs)


# --------------------------
if __name__ == "__main__":

    printm("Testing GPU spectrum calculation:", pytest.main(["test_gpu.py"]))



import ctypes
from os.path import join

import numpy as np
from scipy.constants import N_A, c, h, k

c_cm = 100 * c
c2 = h * c_cm / k

from radis.misc.utils import getProjectRoot


class initData(ctypes.Structure):
    _fields_ = [
        ("v_min", ctypes.c_float),
        ("v_max", ctypes.c_float),
        ("dv", ctypes.c_float),
        ("N_v", ctypes.c_int),
        ("dxG", ctypes.c_float),
        ("dxL", ctypes.c_float),
        ("N_total", ctypes.c_int),
        ("Max_lines", ctypes.c_int),
        ("N_lines", ctypes.c_int),
        ("N_points_per_block", ctypes.c_int),
        ("N_threads_per_block", ctypes.c_int),
        ("N_blocks_per_grid", ctypes.c_int),
        ("N_points_per_thread", ctypes.c_int),
        ("N_iterations_per_thread", ctypes.c_int),
        ("shared_size_floats", ctypes.c_int),
        ("log_c2Mm", ctypes.c_float * 16),
    ]


class iterData(ctypes.Structure):
    _fields_ = [
        ("p", ctypes.c_float),
        ("log_2p", ctypes.c_float),
        ("hlog_T", ctypes.c_float),
        ("log_rT", ctypes.c_float),
        ("c2T", ctypes.c_float),
        ("N", ctypes.c_float),
        ("l", ctypes.c_float),
        ("slit_FWHM", ctypes.c_float),
        ("log_wG_min", ctypes.c_float),
        ("log_wL_min", ctypes.c_float),
        ("N_G", ctypes.c_int),
        ("N_L", ctypes.c_int),
        ("Q", ctypes.c_float * 16),
    ]


init_h = initData()
iter_h = iterData()


def py_calc_lorentzian_envelope_params(na, gamma, verbose=False):
    """


    Parameters
    ----------
    na : TYPE
        DESCRIPTION.
    gamma : TYPE
        DESCRIPTION.
    verbose : TYPE, optional
        DESCRIPTION. The default is False.

    Returns
    -------
    TYPE
        DESCRIPTION..

    """
    # Remove duplicates
    unique_lines = set([])
    for i in range(len(na)):
        unique_lines.add(str(na[i]) + " " + str(gamma[i]))

    # Only keep extremes
    max_dict = {}
    min_dict = {}
    for s in unique_lines:
        na_i, gamma_i = map(float, s.split())
        try:
            min_dict[na_i] = gamma_i if gamma_i < min_dict[na_i] else min_dict[na_i]
            max_dict[na_i] = gamma_i if gamma_i > max_dict[na_i] else max_dict[na_i]

        except (KeyError):
            min_dict[na_i] = gamma_i
            max_dict[na_i] = gamma_i

    # Check which ones are really at the top:
    result = []
    for test_dict in (min_dict, max_dict):

        keys = sorted(test_dict.keys(), reverse=(test_dict == min_dict))
        A = [keys[0]]
        B = [np.log(test_dict[keys[0]])]
        X = [-np.inf]

        for key in keys[1:]:
            for i in range(len(X)):
                xi = (np.log(test_dict[key]) - B[i]) / (A[i] - key)
                if xi >= X[i]:
                    if i < len(X) - 1:
                        if xi < X[i + 1]:
                            break
                    else:
                        break

            A = A[: i + 1] + [key]
            B = B[: i + 1] + [np.log(test_dict[key])]
            X = X[: i + 1] + [xi]

        X = X[1:] + [np.inf]
        result.append((A, B, X))

    return tuple(result)


def py_calc_gaussian_envelope_params(log_2vMm, verbose=False):
    """


    Parameters
    ----------
    log_2vMm : TYPE
        DESCRIPTION.
    verbose : TYPE, optional
        DESCRIPTION. The default is False.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    return np.amin(log_2vMm), np.max(log_2vMm)


try:
    from radis_cython_extensions import (  # isort:skip
        calc_gaussian_envelope_params,
        calc_lorentzian_envelope_params,
    )
except (ModuleNotFoundError):
    calc_gaussian_envelope_params = py_calc_gaussian_envelope_params
    calc_lorentzian_envelope_params = py_calc_lorentzian_envelope_params


def init_gaussian_params(log_2vMm, verbose):
    """


    Parameters
    ----------
    log_2vMm : TYPE
        DESCRIPTION.
    verbose : TYPE
        DESCRIPTION.

    Returns
    -------
    param_data : TYPE
        DESCRIPTION.

    """

    if verbose >= 2:
        print("Initializing Gaussian parameters")

    ##fname = "Gaussian_minmax_" + str(len(log_2vMm)) + ".dat"
    ##    try:
    ##        param_data = pickle.load(open(fname, "rb"))
    ##        if verbose >= 2:
    ##            print(" (from cache)... ")
    ##
    ##    except (OSError, IOError):
    if True:
        if verbose >= 2:
            print("... ")

        param_data = calc_gaussian_envelope_params(log_2vMm, verbose)
    ##        pickle.dump(param_data, open(fname, "wb"))

    if verbose >= 2:
        print("done!")

    return param_data


def init_lorentzian_params(na, gamma, verbose):
    """


    Parameters
    ----------
    na : TYPE
        DESCRIPTION.
    gamma : TYPE
        DESCRIPTION.
    verbose : TYPE
        DESCRIPTION.

    Returns
    -------
    param_data : TYPE
        DESCRIPTION.

    """

    if verbose >= 2:
        print("Initializing Lorentzian parameters ")

    ##fname = "Lorenzian_minmax_" + str(len(gamma)) + ".dat"
    ##
    ##    try:
    ##        with open(fname, "rb") as f:
    ##            if verbose >= 2:
    ##                print(" (from cache)... ")
    ##            param_data = pickle.load(f)
    ##
    ##    except:
    if True:
        if verbose >= 2:
            print(" ... ")

        param_data = calc_lorentzian_envelope_params(na, gamma, verbose)
    ##        with open(fname, "wb") as f:
    ##            pickle.dump(param_data, f)

    if verbose >= 2:
        print("done!")

    return param_data


def calc_gaussian_params(gaussian_param_data, init_h, iter_h, epsilon=1e-4):
    """


    Parameters
    ----------
    gaussian_param_data : TYPE
        DESCRIPTION.
    init_h : TYPE
        DESCRIPTION.
    iter_h : TYPE
        DESCRIPTION.
    epsilon : TYPE, optional
        DESCRIPTION. The default is 1e-4.

    Returns
    -------
    None.

    """

    log_2vMm_min, log_2vMm_max = gaussian_param_data
    log_wG_min = log_2vMm_min + iter_h.hlog_T
    log_wG_max = log_2vMm_max + iter_h.hlog_T
    ##    print("wG:", log_wG_min, log_wG_max)
    log_wG_max += epsilon

    # dxG = (log_wG_max - log_wG_min) / (init_h.N_G - 1)
    N = int(np.ceil((log_wG_max - log_wG_min) / init_h.dxG) + 1)

    iter_h.log_wG_min = log_wG_min
    # iter_h.dxG = dxG
    iter_h.N_G = N


def calc_lorentzian_minmax(param_data, log_rT, log_2p):
    """


    Parameters
    ----------
    param_data : TYPE
        DESCRIPTION.
    log_rT : TYPE
        DESCRIPTION.
    log_2p : TYPE
        DESCRIPTION.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """

    result = []
    for params in param_data:
        A, B, X = params
        i = 0
        while X[i] < log_rT:
            i += 1
        result.append(log_rT * A[i] + B[i] + log_2p)
    return tuple(result)


def calc_lorentzian_params(param_data, init_h, iter_h, epsilon=1e-4):
    """


    Parameters
    ----------
    param_data : TYPE
        DESCRIPTION.
    init_h : TYPE
        DESCRIPTION.
    iter_h : TYPE
        DESCRIPTION.
    epsilon : TYPE, optional
        DESCRIPTION. The default is 1e-4.

    Returns
    -------
    None.

    """

    log_wL_min, log_wL_max = calc_lorentzian_minmax(
        param_data, iter_h.log_rT, iter_h.log_2p
    )
    log_wL_max += epsilon
    # dxL = (log_wL_max - log_wL_min) / (init_h.N_L - 1)
    N = int(np.ceil((log_wL_max - log_wL_min) / init_h.dxL) + 1)

    iter_h.log_wL_min = log_wL_min
    iter_h.N_L = N


def set_pTQ(p, T, mole_fraction, iter_h, l=1.0, slit_FWHM=0.0):
    """


    Parameters
    ----------
    p : float
        pressure [bar].
    T : float
        temperature [K].
    mole_fraction : float
    iter_h : TYPE
        DESCRIPTION.
    l : TYPE, optional
        DESCRIPTION. The default is 1.0.
    slit_FWHM : TYPE, optional
        DESCRIPTION. The default is 0.0.

    Returns
    -------
    None.

    """

    iter_h.p = p  # bar
    iter_h.log_2p = np.log(2 * p)
    iter_h.hlog_T = 0.5 * np.log(T)
    iter_h.log_rT = np.log(296.0 / T)
    iter_h.c2T = -c2 / T
    iter_h.N = mole_fraction * p * 1e5 / (1e6 * k * T)  # cm-3
    iter_h.l = l
    iter_h.slit_FWHM = slit_FWHM

    for i in range(len(Q_interpolator_list)):
        iter_h.Q[i] = Q_interpolator_list[i](T)


def constant_memory_setter(cuda_module, var_str):
    def setter(var_h):
        memptr_d = cuda_module.get_global(var_str)
        ptr = ctypes.cast(ctypes.pointer(var_h), ctypes.c_void_p)
        struct_size = ctypes.sizeof(var_h)
        memptr_d.copy_from_host(ptr, struct_size)

    return setter


def gpu_init(
    v_arr,
    dxG,
    dxL,
    iso,
    v0,
    da,
    gamma,
    na,
    S0,
    El,
    Mm_arr,
    Q_intp_list,
    verbose=True,
    gpu=False,
):
    """


    Parameters
    ----------
    v_arr : TYPE
        DESCRIPTION.
    dxG : TYPE
        DESCRIPTION.
    dxL : TYPE
        DESCRIPTION.
    iso : TYPE
        DESCRIPTION.
    v0 : TYPE
        DESCRIPTION.
    da : TYPE
        DESCRIPTION.
    gamma : TYPE
        DESCRIPTION.
    na : TYPE
        DESCRIPTION.
    S0 : TYPE
        DESCRIPTION.
    El : TYPE
        DESCRIPTION.
    Mm_arr  : TYPE
        DESCRIPTION.
    Q_intp_list: TYPE
        DESCRIPTION.
    verbose : TYPE, optional
        DESCRIPTION. The default is True.
    gpu : TYPE, optional
        DESCRIPTION. The default is False.

    Returns
    -------
    None.

    """

    # ----------- setup global variables -----------------
    global init_h
    global block_preparation_step_size
    global iso_d, v0_d, da_d, S0_d, El_d, gamma_d, na_d
    global spectrum_in_d, transmittance_noslit_d, transmittance_FT_d
    global lorentzian_param_data, gaussian_param_data, Q_interpolator_list
    global cuda_module, cuda_functions
    # -----------------------------------------------------

    if gpu:
        from cupy import RawModule, array, complex64, float32, zeros

        cuda_fname = join(getProjectRoot(), "lbl", "gpu.cpp")
        with open(cuda_fname, "rb") as f:
            cuda_code = f.read().decode()

        cuda_module = RawModule(code=cuda_code)
        cuda_functions = (
            cuda_module.get_function("fillLDM"),
            cuda_module.get_function("applyLineshapes"),
            cuda_module.get_function("calcTransmittanceNoslit"),
            cuda_module.get_function("applyGaussianSlit"),
        )

        set_init_params = constant_memory_setter(cuda_module, "init_d")

    else:
        from numpy import complex64, float32, zeros

        array = lambda arr: arr
        from radis_cython_extensions import set_init_params  # isort:skip

    init_h.v_min = np.min(v_arr)  # 2000.0
    init_h.v_max = np.max(v_arr)  # 2400.0
    init_h.dv = (v_arr[-1] - v_arr[0]) / (len(v_arr) - 1)  # 0.002
    init_h.N_v = len(v_arr)
    init_h.dxG = dxG
    init_h.dxL = dxL

    init_h.N_iterations_per_thread = 1024
    block_preparation_step_size = 128

    shared_size = 0x8000  # Bytes - Size of the shared memory
    init_h.shared_size_floats = shared_size // 4  # size of float

    # init_h.N_total = init_h.N_v * init_h.N_G * init_h.N_L
    # init_h.N_points_per_block = init_h.shared_size_floats // (init_h.N_v * init_h.N_G)

    init_h.N_threads_per_block = 1024
    init_h.N_blocks_per_grid = 4 * 256 * 256
    init_h.N_points_per_thread = init_h.N_points_per_block // init_h.N_threads_per_block

    if verbose >= 2:
        print()
        print("Spectral points per block  : {0}".format(init_h.N_points_per_block))
        print("Threads per block          : {0}".format(init_h.N_threads_per_block))
        print("Spectral points per thread : {0}".format(init_h.N_points_per_thread))
        print()

    init_h.N_lines = int(len(v0))

    if verbose == 1:
        print("Number of lines loaded: {0}".format(init_h.N_lines))
        print()

    if verbose >= 2:
        print("Allocating device memory and copying data...")

    # S_klm_d = zeros(
    #     (2 * init_h.N_v, init_h.N_G, init_h.N_L),
    #     order="C",
    #     dtype=float32,
    # )
    spectrum_in_d = zeros(init_h.N_v + 1, dtype=complex64)
    transmittance_noslit_d = zeros(init_h.N_v * 2, dtype=float32)
    transmittance_FT_d = zeros(init_h.N_v + 1, dtype=complex64)

    # #Copy spectral data to device
    iso_d = array(iso)
    v0_d = array(v0)
    da_d = array(da)
    S0_d = array(S0)
    El_d = array(El)
    gamma_d = array(gamma)
    na_d = array(na)

    if verbose >= 2:
        print("done!")

    log_c2Mm_arr = np.array(
        [0]
        + [
            0.5 * np.log(8 * k * np.log(2) / (c**2 * Mm * 1e-3 / N_A))
            for Mm in Mm_arr[1:]
        ]
    )
    for i in range(len(log_c2Mm_arr)):
        init_h.log_c2Mm[i] = log_c2Mm_arr[i]
    Q_interpolator_list = Q_intp_list

    log_2vMm = np.log(v0) + log_c2Mm_arr.take(iso)
    gaussian_param_data = init_gaussian_params(log_2vMm.astype(np.float32), verbose)
    lorentzian_param_data = init_lorentzian_params(na, gamma, verbose)

    if verbose >= 2:
        print("Copying initialization parameters to device memory...")
    set_init_params(init_h)

    if verbose >= 2:
        print("done!")
        print("Copying spectral data to device memory...")


##    print("wG fast:",np.min(log_2vMm),np.max(log_2vMm))
##
##    log_wG_debug_h = np.log(v0) + 0.5*np.log(8*k*np.log(2)/(c**2*Mm_arr[iso]))
##    print("wG arr:",np.min(log_wG_debug_h),np.max(log_wG_debug_h))


def gpu_iterate(p, T, mole_fraction, l=1.0, slit_FWHM=0.0, verbose=0, gpu=False):
    """
    Parameters
    ----------
    p : float
        pressure [bar]
    T : float
        temperature [K]
    mole_fraction : float

    Other Parameters
    ----------------
    verbose : bool, optional
        The default is True.
    l : TYPE, optional
        DESCRIPTION. The default is 1.0.
    slit_FWHM : TYPE, optional
        DESCRIPTION. The default is 0.0.
    gpu : TYPE, optional
        DESCRIPTION. The default is False.

    Returns
    -------
    abscoeff_h : TYPE
        DESCRIPTION.
    transmittance_h : TYPE
        DESCRIPTION.

    """

    # ----------- setup global variables -----------------
    global init_h, iter_h
    global iso_d, v0_d, da_d, S0_d, El_d, gamma_d, na_d
    global spectrum_in_d, transmittance_noslit_d, transmittance_FT_d
    global cuda_module, cuda_functions
    global start, stop, elapsedTime
    # ------------------------------------------------------

    if gpu:
        from cupy import asnumpy, complex64, float32, zeros
        from cupy.cuda.runtime import deviceSynchronize
        from cupy.fft import irfft, rfft

        (
            fillLDM,
            applyLineshapes,
            calcTransmittanceNoslit,
            applyGaussianSlit,
        ) = cuda_functions
        set_iter_params = constant_memory_setter(cuda_module, "iter_d")

    else:
        # isort: off
        from numpy import complex64, float32, zeros
        from numpy.fft import irfft, rfft

        from radis_cython_extensions import (  # isort:skip
            applyGaussianSlit,
            applyLineshapes,
            calcTransmittanceNoslit,
            fillLDM,
            set_iter_params,
        )

        # isort: on

        asnumpy = lambda arr: arr

    if verbose >= 2:
        print("Copying iteration parameters to device...")

    set_pTQ(p, T, mole_fraction, iter_h, l=l, slit_FWHM=slit_FWHM)

    calc_gaussian_params(
        gaussian_param_data,
        init_h,
        iter_h,
    )

    calc_lorentzian_params(
        lorentzian_param_data,
        init_h,
        iter_h,
    )

    set_iter_params(iter_h)

    if verbose >= 2:
        print("done!")
        print("Filling LDM...")

    # S_klm_d.fill(0)

    S_klm_d = zeros(
        (2 * init_h.N_v, iter_h.N_G, iter_h.N_L),
        order="C",
        dtype=float32,
    )

    spectrum_in_d.fill(0)
    transmittance_FT_d.fill(0)

    n_threads = init_h.N_threads_per_block
    n_blocks = init_h.N_lines // (n_threads * init_h.N_iterations_per_thread) + 1

    fillLDM(
        (n_blocks,),
        (n_threads,),
        (
            iso_d,
            v0_d,
            da_d,
            S0_d,
            El_d,
            gamma_d,
            na_d,
            S_klm_d,
        ),
    )
    if gpu:
        deviceSynchronize()

    if verbose >= 2:
        print("Applying lineshapes...")

    S_klm_FT_d = rfft(S_klm_d, axis=0).astype(complex64)

    if gpu:
        deviceSynchronize()

    n_threads = init_h.N_threads_per_block
    n_blocks = (init_h.N_v + 1) // n_threads + 1
    applyLineshapes(
        (n_blocks,),
        (n_threads,),
        (
            S_klm_FT_d,
            spectrum_in_d,
        ),
    )

    if gpu:
        deviceSynchronize()

    spectrum_out_d = irfft(spectrum_in_d).astype(float32)

    if gpu:
        deviceSynchronize()

    if verbose >= 2:
        print("Done!")
        print("Calculating transmittance...")

    abscoeff_h = asnumpy(spectrum_out_d)[: init_h.N_v]

    ## Calc transmittance_noslit:
    n_threads = init_h.N_threads_per_block
    n_blocks = (2 * init_h.N_v) // n_threads + 1
    calcTransmittanceNoslit(
        (n_blocks,),
        (n_threads,),
        (spectrum_out_d, transmittance_noslit_d),
    )

    if gpu:
        deviceSynchronize()

    if verbose >= 2:
        print("Done!")
        print("Applying slit function...")

    ## Apply slit function:
    transmittance_noslit_FT_d = rfft(transmittance_noslit_d).astype(complex64)

    if gpu:
        deviceSynchronize()

    n_threads = init_h.N_threads_per_block
    n_blocks = (init_h.N_v + 1) // n_threads + 1
    applyGaussianSlit(
        (n_blocks,),
        (n_threads,),
        (transmittance_noslit_FT_d, transmittance_FT_d),
    )

    if gpu:
        deviceSynchronize()

    transmittance_d = irfft(transmittance_FT_d).astype(float32)

    if gpu:
        deviceSynchronize()

    transmittance_h = asnumpy(transmittance_d)[: init_h.N_v]

    if verbose >= 2:
        print("done!")

    if verbose == 1:
        print("Finished calculating spectrum!")

    return abscoeff_h, transmittance_h, iter_h



# -*- coding: utf-8 -*-
"""

Module for New RADIS Fitting Interface

This module is to conduct benchmarking process on spectra (LTE or Non-LTE) with
temperatures (gas, vibrational, rotational, translational), mole fraction, cutoff,
as well as other characteristics as fitting parameters.

This module is implemented into RADIS from the Fitting Benchmarking repository, as
a 2022 GSOC project: https://github.com/TranHuuNhatHuy/RADIS-Spectrum-Fitting-Benchmark

This module aims to provide RADIS with a unified fitting interface which covers most of
current fitting cases encountered from fellow scientists and researchers, while being
simple and easy-to-use to RADIS newcomers, as you only need to input the parameters into
pre-defined Dictionary objects in a simplified script, just as filling a form. This also
serves as a bridge between RADIS codebase and a future fitting feature on RADIS App.

"""


import json
import time
from typing import Union

import numpy as np
from lmfit import Parameters, fit_report, minimize
from lmfit.minimizer import MinimizerResult

from radis import Spectrum, SpectrumFactory
from radis.spectrum import plot_diff
from radis.spectrum.compare import get_residual
from radis.tools.database import load_spec

# --------------------- MODELS / COST FUNCTIONS --------------------- #


def residual_LTE(params, conditions, s_data, sf, log, verbose):
    """A cost function that calculates an LTE spectrum based on the
    initial conditions and values of fit parameters, then returning a
    scalar containing difference between experimental and data spectra.

    Parameters
    ----------
    params : LMFIT.Parameters
        basically a dict containing fit parameters that will be varied
        during the minimization.
    conditions: dict
        a dict containing conditions (fixed parameters) for generating
        model spectra.
    s_data: Spectrum
        data spectrum, loaded from the directory stated in JSON file.
    sf: SpectrumFactory
        the SpectrumFactory object used for modeling spectra, generated
        before the minimize loop occurs.
    log: list
        a Dictionary storing runtime log of the fitting process that are
        not quite covered by the Minimizer, including: residual and fit
        values after each fitting loop, and total time elapsed.

    Other parameters
    ----------------
    verbose : bool
        if True, will print out result of each fitting loop.

    Returns
    -------
    residual: float
        residuals of the two spectra, using RADIS's get_residual().

    Examples
    --------

    .. minigallery:: radis.tools.new_fitting.residual_LTE

    See Also
    --------
    :py:func:`~radis.tools.new_fitting.residual_NonLTE`
    :py:func:`~radis.tools.fitting.LTEModel`

    """

    # GENERATE LTE SPECTRUM BASED ON THOSE PARAMETERS

    # Load initial values of fit parameters

    # These keys must be ignored when parsing parameters. They will be dealt with separately
    ignore_keys = [
        "offsetnm",
        "offsetcm1",
    ]

    # Retrieve current values of parameters, for calculating model spectrum
    kwargs = {}
    for param in params:
        if param not in ignore_keys:
            kwargs[param] = float(params[param])

    # Model spectrum calculation
    s_model = sf.eq_spectrum(**kwargs)

    # Deal with "offset"
    if "offsetnm" in params:
        offset_value = float(params["offsetnm"])
        s_model = s_model.offset(offset_value, "nm")
    if "offsetcm1" in params:
        offset_value = float(params["offsetcm1"])
        s_model = s_model.offset(offset_value, "cm-1")

    # FURTHER REFINE THE MODELED SPECTRUM BEFORE CALCULATING DIFF

    pipeline = conditions["pipeline"]
    model = conditions["model"]

    # Apply slit stated in "model"
    if "slit" in model:

        slit_model = model["slit"]

        # The user uses simple format of "[value] [unit]", such as "-0.2 nm"
        if isinstance(slit_model, str):
            slit_val, slit_unit = slit_model.split()
            s_model = s_model.apply_slit(float(slit_val), slit_unit)

        # The user uses a dict with complex format of slit function, unit, shape, center wavespace, dispersion, etc.
        if isinstance(slit_model, dict):
            kwargs = {}
            for cond in slit_model:
                kwargs[cond] = slit_model[cond]
            s_model = s_model.apply_slit(**kwargs)

    # Take spectral quantity
    fit_var = pipeline["fit_var"]
    s_model = s_model.take(fit_var)

    # Apply offset stated in "model"
    if "offset" in model:
        off_val, off_unit = model["offset"].split()
        s_model = s_model.offset(float(off_val), off_unit)

    # Apply normalization
    if "normalize" in pipeline:
        if pipeline["normalize"]:
            s_model = s_model.normalize()

    # ACQUIRE AND RETURN DIFF, ALSO LOG FITTING HISTORY

    # Acquire diff
    residual = get_residual(s_model, s_data, fit_var, norm="L2", ignore_nan="True")

    # Log the current residual
    log["residual"].append(residual)

    # Log the current fitting values of fit parameters
    current_fitvals = []
    for param in params:
        current_fitvals.append(float(params[param]))
    log["fit_vals"].append(current_fitvals)

    return residual


def residual_NonLTE(params, conditions, s_data, sf, log, verbose):
    """A cost function that calculates a non-LTE spectrum based on the
    initial conditions and values of fit parameters, then returning a
    scalar containing difference between experimental and data spectra.

    Parameters
    ----------
    params : LMFIT.Parameters
        basically a dict containing fit parameters that will be varied
        during the minimization.
    conditions: dict
        a dict containing conditions (fixed parameters) for generating
        model spectra.
    s_data: Spectrum
        data spectrum, loaded from the directory stated in JSON file.
    sf: SpectrumFactory
        the SpectrumFactory object used for modeling spectra, generated
        before the minimize loop occurs.
    log: list
        a Dictionary storing runtime log of the fitting process that are
        not quite covered by the Minimizer, including: residual and fit
        values after each fitting loop, and total time elapsed.

    Other parameters
    ----------------
    verbose : bool
        if True, will print out result of each fitting loop.

    Returns
    -------
    residual: float
        residuals of the two spectra, using RADIS's get_residual().

    Examples
    --------

    .. minigallery:: radis.tools.new_fitting.residual_NonLTE

    See Also
    --------
    :py:func:`~radis.tools.new_fitting.residual_LTE`
    :py:func:`~radis.tools.fitting.radis.tools.fitting.Tvib12Tvib3Trot_NonLTEModel`

    """

    # GENERATE NON-LTE SPECTRUM BASED ON THOSE PARAMETERS

    # Load initial values of fit parameters

    # These keys must be ignored when parsing parameters. They will be dealt with separately
    ignore_keys = [
        "offsetnm",
        "offsetcm-1",
    ]

    # Retrieve current values of parameters, for calculating model spectrum
    kwargs = {}
    for param in params:
        if param not in ignore_keys:
            kwargs[param] = float(params[param])

    # Deal with the case of multiple Tvib temperatures
    if "Tvib0" in kwargs:  # There is trace of a "Tvib fragmentation" before
        Tvib = []
        for kw in kwargs:
            if "Tvib" in kw:  # Such as "Tvib0", "Tvib1" or so
                Tvib.append(kwargs[kw])  # Bring them altogether, uwu
                kwargs.pop(kw)  # Dispose the fragmented one in kwargs
        kwargs["Tvib"] = tuple(Tvib)  # Finally, we have the tuple of Tvib

    # Model spectrum calculation
    s_model = sf.non_eq_spectrum(**kwargs)

    # Deal with "offset"
    if "offsetnm" in params:
        offset_value = float(params["offsetnm"])
        s_model = s_model.offset(offset_value, "nm")
    if "offsetcm-1" in params:
        offset_value = float(params["offsetcm-1"])
        s_model = s_model.offset(offset_value, "cm-1")

    # FURTHER REFINE THE MODELED SPECTRUM BEFORE CALCULATING DIFF

    pipeline = conditions["pipeline"]
    model = conditions["model"]

    # Apply slit stated in "model"
    if "slit" in model:

        slit_model = model["slit"]

        # The user uses simple format of "[value] [unit]", such as "-0.2 nm"
        if isinstance(slit_model, str):
            slit_val, slit_unit = slit_model.split()
            s_model = s_model.apply_slit(float(slit_val), slit_unit)

        # The user uses a dict with complex format of slit function, unit, shape, center wavespace, dispersion, etc.
        if isinstance(slit_model, dict):
            kwargs = {}
            for cond in slit_model:
                kwargs[cond] = slit_model[cond]
            s_model = s_model.apply_slit(**kwargs)

    # Take spectral quantity
    fit_var = pipeline["fit_var"]
    s_model = s_model.take(fit_var)

    # Apply offset
    if "offset" in model:
        off_val, off_unit = model["offset"].split()
        s_model = s_model.offset(float(off_val), off_unit)

    # Apply normalization
    if "normalize" in pipeline:
        if pipeline["normalize"]:
            s_model = s_model.normalize()

    # ACQUIRE AND RETURN DIFF, ALSO LOG FITTING HISTORY

    # Acquire diff
    residual = get_residual(s_data, s_model, fit_var, norm="L2", ignore_nan="True")

    # Log the current residual
    log["residual"].append(residual)

    # Log the current fitting values of fit parameters
    current_fitvals = []
    for param in params:
        current_fitvals.append(float(params[param]))
    log["fit_vals"].append(current_fitvals)

    return residual


# ----------------------------------------------------------------------- #


# --------------------- FITTING METHODS / FUNCTIONS --------------------- #


def get_conditions(
    input_conditions,
    verbose=True,
) -> Union[dict, Parameters]:
    """Get the ground-truth information by reading from input - either a JSON file
    or a dict parameter. Return a dict parameter containing fixed conditions for
    spectrum calculation, and an LMFIT.Parameters object containing fit parameters.

    Parameters
    ----------
    input_conditions : str or dict
        If str, the code assumes this is path to the JSON input file from the script
        calling it, for example:: "./test_dir/CO2_measured_spectrum_4-5um.json", and
        then parses the JSON data as a dict.
        If dict, the code just simply parses the dict, yeah, that simple.

    Other Parameters
    ----------------
    verbose: bool
        by default, True, print details about the fitting progress.

    Examples
    --------

    .. minigallery:: radis.tools.new_fitting.get_conditions

    """

    # LOG THE INPUT CONDITIONS (JSON OR DICT)

    if type(input_conditions) == str:  # Must be a JSON path

        with open(input_conditions, "r") as f:  # Load JSON as dict
            conditions = json.load(f)
            f.close()

    elif type(input_conditions) == dict:  # Must be just a dict

        conditions = input_conditions

    else:

        raise TypeError(
            f"Unrecognized input. Got type {type(input_conditions)}, expecting str or dict."
        )

    # READ "fit" KEY AND LOAD IT INTO NEWLY INITIALIZED PARAMETERS OBJECT

    params = Parameters()  # Initiate Parameters object

    fit_params = conditions.pop("fit")
    dhb = 1000  # Default half-bound for fit temperatures, in case user don't specify bounds

    # Deal with "Tvib", especially if it has multiple temperatures

    if "Tvib" in fit_params:

        Tvib = fit_params["Tvib"]

        if (type(Tvib) == list) or (
            type(Tvib) == tuple
        ):  # Multiple temperatures detected

            # Split Tvib[] into Tvib0, Tvib1, Tvib2,... for later being put into Parameters
            for i in range(len(Tvib)):
                fit_params[f"Tvib{i}"] = Tvib[i]

            # Destroy the original "Tvib" key
            fit_params.pop("Tvib")

    # Deal with "offset", which is something like "-0.2 nm", as requested by Mr. Minou

    if "offset" in fit_params:

        offset_value, offset_unit = fit_params.pop("offset").split()
        if offset_unit == "cm-1":
            offset_unitName = "cm1"
        elif offset_unit == "nm":
            offset_unitName = "nm"
        else:
            raise ValueError(
                f'Unrecognized offset unit {offset_unit}. Only "nm" and "cm-1" are accepted.'
            )

        # The reason for all the mess above is, LMFIT does not accept parameter name with hyphen like cm-1.
        # Thus, I remove the hyphen when assigning its name into LMFIT Parameters object.

        # Get numerical value only, and add unit to name for later identification
        fit_params[f"offset{offset_unitName}"] = float(offset_value)

    # Put every fit parameter into Parameters, with the default bounding range

    for param in fit_params:

        init_val = fit_params[param]

        if param == "mole_fraction":
            init_bound = [0, 1]
        elif "offset" in param:  # Either "offsetnm" or "offsetcm1"
            init_bound = [-1, 1]
        else:
            init_bound = [(init_val - dhb) if init_val > dhb else 0, init_val + dhb]

        params.add(
            param,
            value=init_val,
            min=init_bound[0],
            max=init_bound[1],
        )

    # READ "bounds" KEY AND RE-ASSIGN DEFAULT BOUNDING RANGES WITH USER-DEFINED ONES

    if "bounds" in conditions:

        bounds = conditions.pop("bounds")

        # Deal with "Tvib", especially if it has multiple temperature-bounding ranges

        if "Tvib" in bounds:

            bound_Tvib = bounds["Tvib"]

            if (
                len(bound_Tvib) > 1
            ):  # Not sure if it's multiple ranges, or just 1 range with 2 ends
                if type(bound_Tvib[0]) in [
                    list,
                    tuple,
                ]:  # Now we sure it's multiple ranges

                    # Split Tvib[] into Tvib0, Tvib1, Tvib2,... for later being put into Parameters
                    for i in range(len(bound_Tvib)):
                        bounds[f"Tvib{i}"] = bound_Tvib[i]

                    # Destroy the original key
                    bounds.pop("Tvib")

        # Deal with "offset", especially now the offset parameter is either "offsetnm" or "offsetcm1"

        if "offset" in bounds:

            bound_offset = bounds.pop("offset")

            if "offsetnm" in params:
                bounds["offsetnm"] = bound_offset

            if "offsetcm1" in params:
                bounds["offsetcm1"] = bound_offset

        # Set user-defined bounding ranges

        for bound_param in bounds:

            bound_val = bounds[bound_param]

            params[bound_param].set(min=bound_val[0], max=bound_val[1])

    return conditions, params


# -------------------------------------------------------------------------------------------- #


def spectrum_refinement(s_data, conditions, verbose) -> Union[Spectrum, dict]:
    """Receive an experimental spectrum and further refine it according to the provided pipeline
    and ground-truths. Refinement process includes extracting the desired spectrum quantity,
    removing NaN values, and other additional convolutions. Finally, a refined Spectrum object is
    returned, along with the original condition as a Dictionary object.

    Parameters
    ----------
    s_data : Spectrum
        experimental spectrum loaded from the path given in input JSON file.
    conditions : dict
        a Dictionary containing all ground-truth information, and also spectrum refinements or
        convolutions (desired spectrum quantity, normalization, etc.), as well as fitting process
        (max loop allowed, terminal tolerance, etc.).

    Other parameters
    ----------------
    verbose : bool
        by default, True, print details about the refinement progress.

    Returns
    -------
    s_refined : Spectrum
        the refined spectrum.
    conditions : dict
        the input conditions that might have been added fit_var (in case user didn't).

    Examples
    --------

    .. minigallery:: radis.tools.new_fitting.spectrum_refinement

    """

    # Extract spectrum

    pipeline = conditions["pipeline"]

    if "fit_var" in pipeline:
        fit_var = pipeline["fit_var"]  # Acquire the stated quantity
    else:
        fit_var = s_data.get_vars()[0]  # If not stated, take first quantity
        conditions["pipeline"]["fit_var"] = fit_var  # And add to the dict

    s_data_wav, s_data_val = s_data.get(fit_var)

    if verbose:
        print(f"\nAcquired spectral quantity '{fit_var}' from the spectrum.")

    # Remove NaN values. A wise man once said, "Nan is good but only in India"

    s_data_mtr = np.vstack(
        (s_data_wav, s_data_val)
    )  # Merge wavelength/wavenumber and intensity into pairs
    s_data_mtr = s_data_mtr[:, ~np.isnan(s_data_mtr).any(axis=0)]  # Purge NaN pairs

    if verbose:
        print(f"NaN values successfully purged.", end="")
        print(f"Number of data points left: {len(s_data_mtr[0])} points.")

    # Recreate the data spectrum with the spectral quantity

    s_refined = (
        Spectrum(
            {fit_var: (s_data_mtr[0], s_data_mtr[1])},
            wunit=s_data.get_waveunit(),
            units={fit_var: s_data.units[fit_var]},
        )
        .take(fit_var)
        .sort()
    )

    # Further refinement

    # Apply normalization if required
    if "normalize" in pipeline:
        if pipeline["normalize"]:
            s_refined = s_refined.normalize()
            if verbose:
                print("Normalization applied.")

    # Currently supporting normalization. May add more in the future if requested.

    return s_refined, conditions


# -------------------------------------------------------------------------------------------- #


def fit_spectrum(
    s_exp=None,
    fit_params=None,
    model=None,
    pipeline=None,
    bounds=None,
    input_file=None,
    verbose=True,
    show_plot=True,
    fit_kws={},
) -> Union[Spectrum, MinimizerResult, dict]:
    """Fit an experimental spectrum (from here referred as "data spectrum") with a modeled one,
    then derive the fit results. Data spectrum is loaded from the path stated in the JSON file,
    while model spectrum is generated based on the conditions stated in the JSON file, too.

    Parameters
    ----------
    s_exp : Spectrum
        a Spectrum object, containing the experimental spectrum.
    fit_params : dict
        a dict object, containing fit parameters and their initial values.
    bounds : dict
        a dict object, optional, containing bounding ranges of those fit parameters. If not stated,
        all fit parameters will receive default bounding ranges.
    model : dict
        a dict object, containing experimental conditions and some additional convolutions for
        modeling the spectrum.
    pipeline : dict
        a dict object, containing some preference properties of fitting process.
    input_file : str
        path to the JSON input file from the script calling it, by default, None. If this parameter
        is defined, it means user uses JSON file for inputing everything, and so will discard all
        other input parameters. For example:: "./test_dir/CO2_measured_spectrum_4-5um.json"

    Other parameters
    ----------
    verbose : bool
        print details about each loop of the fitting process. By default, False.
    show_plot : bool
        show experimental spectrum before and after being fitted. By default, True. Showing the plot
        will pause the runtime, so if you are running benchmark which involves multiple fitting cases,
        set it to False.
    fit_kws: dict
        Options to pass to the minimizer being used. See :py:func:`~lmfit.minimizer.minimize`

    Returns
    -------
    s_best: Spectrum
        visualization of the best fit results obtained, as a spectrum.
    best_fit: MinimizerResult
        best fit results, output of LMFIT MinimizeResult.
    log: dict
        a Dictionary storing runtime log of the fitting process that are
        not quite covered by the Minimizer, including: residual and fit
        values after each fitting loop, and total time elapsed.

    Examples
    --------

    ::
        from radis import load_spec
        from radis.test.utils import getTestFile
        from radis.tools.new_fitting import fit_spectrum

        # Load an experimental spectrum. You can prepare yours, or fetch one of them in the radis/test/files directory.
        my_spec = getTestFile("synth-NH3-1-500-2000cm-P10-mf0.01-p1.spec")
        s_experimental = load_spec(my_spec)

        # Experimental conditions which will be used for spectrum modeling. Basically, these are known ground-truths.
        experimental_conditions = {
            "molecule": "NH3",  # Molecule ID
            "isotope": "1",  # Isotope ID, can have multiple at once
            "wmin": 1000,  # Starting wavelength/wavenumber to be cropped out from the original experimental spectrum.
            "wmax": 1050,  # Ending wavelength/wavenumber for the cropping range.
            "wunit": "cm-1",  # Accompanying unit of those 2 wavelengths/wavenumbers above.
            "mole_fraction": 0.01,  # Species mole fraction, from 0 to 1.
            "pressure": 10,  # Total pressure of gas, in "bar" unit by default, but you can also use Astropy units.
            "path_length": 1,  # Experimental path length, in "cm" unit by default, but you can also use Astropy units.
            "slit": "1 nm",  # Experimental slit, must be a blank space separating slit amount and unit.
            "offset": "-0.2 nm",  # Experimental offset, must be a blank space separating offset amount and unit.
            "wstep": 0.003,  # Resolution of wavenumber grid, in cm-1.
            "databank": "hitran",  # Databank used for the spectrum calculation. Must be stated.
        }

        # List of parameters to be fitted, accompanied by their initial values.
        fit_parameters = {
            "Tgas": 700,  # Gas temperature, in K.
        }

        # List of bounding ranges applied for those fit parameters above.
        bounding_ranges = {
            "Tgas": [
                500,
                2000,
            ],
        }

        # Fitting pipeline setups.
        fit_properties = {
            "method": "leastsq",  # Preferred fitting method. By default, "leastsq".
            "fit_var": "radiance",  # Spectral quantity to be extracted for fitting process, such as "radiance", "absorbance", etc.
            "normalize": False,  # Either applying normalization on both spectra or not.
            "max_loop": 150,  # Max number of loops allowed. By default, 200.
            "tol": 1e-15,  # Fitting tolerance, only applicable for "lbfgsb" method.
        }

        # Conduct the fitting process!
        s_best, result, log = fit_spectrum(
            s_exp=s_experimental,  # Experimental spectrum.
            fit_params=fit_parameters,  # Fit parameters.
            bounds=bounding_ranges,  # Bounding ranges for those fit parameters.
            model=experimental_conditions,  # Experimental ground-truths conditions.
            pipeline=fit_properties,  # Fitting pipeline references.
            verbose=False,  # If you want a clean result, stay False. If you want to see more about each loop, True.
        )
    .. minigallery:: radis.tools.new_fitting.fit_spectrum
    Notes
    -----
    For the fitting method, you can try one among 17 different fitting methods and algorithms of LMFIT, introduced
    in `LMFIT method list <https://lmfit.github.io/lmfit-py/fitting.html#choosing-different-fitting-methods>`.
    See Also
    --------
    :py:func:`~radis.tools.fitting.fit_spectrum()`,
    :py:meth:`~radis.lbl.factory.SpectrumFactory.fit_spectrum`,
    :py:mod:`fitroom`

    """

    begin = time.time()  # Start the fitting time counter

    if verbose:
        print(
            "\n======================= COMMENCE FITTING PROCESS ======================="
        )

    # ACQUIRE AND REFINE EXPERIMENTAL SPECTRUM s_data

    if input_file != None:  # The user uses JSON file

        # Load data from JSON file, then create a Parameters object
        conditions, params = get_conditions(input_file)

        # Get s_data spectrum from the path stated in acquired JSON data, assuming the JSON
        # and the .spec files are in the same directory (it SHOULD be!)

        fileName = conditions["model"].pop("fileName")
        if fileName[-5:] != ".spec":  # fileName doesn't have ".spec" extension
            raise ValueError('Warning: fileName must include ".spec" extension!')

        input_dir = "/".join(input_file.split("/")[0:-1]) + "/"
        spec_path = input_dir + fileName

        # Load and crop the experimental spectrum
        s_data = load_spec(spec_path).crop(
            conditions["model"]["wmin"],
            conditions["model"]["wmax"],
            conditions["model"]["wunit"],
        )

    else:  # The user states a bunch of dicts separately

        # Merge all separated dict inputs into a single dict first

        conditions = {
            "model": model,
            "fit": fit_params,
            "bounds": bounds if bounds is not None else {},
            "pipeline": pipeline,
        }

        # Load data, create Parameters object, from the above dict
        conditions, params = get_conditions(conditions)

        # Load and crop the experimental spectrum
        if "wunit" in conditions["model"]:  # Users state wunit directly
            s_data = s_exp.crop(
                conditions["model"]["wmin"],
                conditions["model"]["wmax"],
                conditions["model"]["wunit"],
            )
        else:  # Users don't state wunit directly, probably astropy units
            s_data = s_exp.crop(
                conditions["model"]["wmin"],
                conditions["model"]["wmax"],
            )

    # Log the time mark when experimental data is retrieved
    end_exp_load = time.time()
    time_exp_load = end_exp_load - begin

    if verbose:
        print(f"\nSuccessfully retrieved the experimental data in {time_exp_load}s.")

    # Further refine the data spectrum before calculating diff

    s_data, conditions = spectrum_refinement(s_data, conditions, verbose)

    pipeline = conditions["pipeline"]
    fit_var = pipeline["fit_var"]

    # Log the time mark when spectrum refinement is over
    end_exp_refine = time.time()
    time_exp_refine = end_exp_refine - end_exp_load

    if verbose:
        print(f"Successfully refined the experimental data in {time_exp_refine}s.")
    if show_plot:
        s_data.plot(show=True)

    # PRE-MINIMIZATION SETUP

    # Create a Spectrum Factory object for modeling spectra

    kwargs = {}
    model = conditions["model"]

    # List of keys that are included in "model", but not included in SpectrumFactory syntax
    ignore_keys = [
        "slit",
        "offset",
        "databank",
    ]

    for cond in model:
        if cond not in ignore_keys:
            kwargs[cond] = model[cond]

    sf = SpectrumFactory(
        **kwargs,
        verbose=False,
        warnings={
            "MissingSelfBroadeningWarning": "ignore",
            "UserWarning": "ignore",
            "NegativeEnergiesWarning": "ignore",
            "HighTemperatureWarning": "ignore",
        },
    )

    # Decide the type of model - LTE or Non-LTE
    LTE = True  # LTE == True means it's LTE
    if ("Tvib" in params) or ("Tvib0" in params):
        LTE = False  # LTE == False means it's non-LTE

    # Determine fitting method, either stated by user or "lbfgsb" by default
    if "method" in pipeline:
        method = pipeline["method"]  # User-defined
    else:
        method = "leastsq"  # By default

    # Determine additional fitting conditions for the minimizer

    fit_kws.update(
        {
            "max_nfev": pipeline["max_loop"]
            if "max_loop" in pipeline
            else 200,  # Default max number of loops: 200
        }
    )

    if "tol" in pipeline:
        if pipeline["method"] == "lbfgsb":
            fit_kws["tol"] = pipeline["tol"]
        else:
            print('"tol" parameter spotted but "method" is not "lbfgsb"!')

    # Prepare fitting log
    log = {"residual": [], "fit_vals": [], "time_fitting": 0}

    # COMMENCE THE FITTING PROCESS

    databank = model.pop("databank")

    # Log the time mark when fitting begins
    begin_fitting = time.time()

    # For LTE spectra
    if LTE:

        sf.fetch_databank(databank, load_columns="equilibrium")

        if verbose:
            print("\nCommence fitting process for LTE spectrum!\n")

        result = minimize(
            residual_LTE,
            params,
            method=method,
            args=[conditions, s_data, sf, log, verbose],
            **fit_kws,
        )

    # For non-LTE spectra
    if not (LTE):

        sf.fetch_databank(databank, load_columns="noneq")

        if verbose:
            print("\nCommence fitting process for non-LTE spectrum!")

        result = minimize(
            residual_NonLTE,
            params,
            method=method,
            args=[conditions, s_data, sf, log, verbose],
            **fit_kws,
        )

    # POST-MINIMIZATION REPORT

    print(fit_report(result), end="\n")  # Report the fitting result

    # REGENERATE THE BEST-FIT SPECTRUM, AS A RESULT FROM THE MINIMIZER

    # Load initial values of fit parameters
    fit_show = {}
    offset_unit = ""
    for name, param in result.params.items():
        if "offset" in name:  # "offset" detected
            offset_value = float(param.value)
            offset_unit = name[6:]
            # Deal with "cm1"
            if offset_unit == "cm1":
                offset_unit = "cm-1"
        else:  # Other parameters not related to offset
            fit_show[name] = float(param.value)
    fit_show["name"] = "best_fit"

    # Generate best fitted spectrum result
    if LTE:
        s_result = sf.eq_spectrum(**fit_show)
    else:
        s_result = sf.non_eq_spectrum(**fit_show)

    # In case offset is detected in parameters
    if offset_unit != "":
        s_result = s_result.offset(offset_value, offset_unit)

    # Apply slit stated in "model"
    if "slit" in model:

        slit_model = model["slit"]

        # The user uses simple format of "[value] [unit]", such as "-0.2 nm"
        if isinstance(slit_model, str):
            slit_val, slit_unit = slit_model.split()
            s_result = s_result.apply_slit(float(slit_val), slit_unit)

        # The user uses a dict with complex format of slit function, unit, shape, center wavespace, dispersion, etc.
        if isinstance(slit_model, dict):
            kwargs = {}
            for cond in slit_model:
                kwargs[cond] = slit_model[cond]
            s_result = s_result.apply_slit(**kwargs)

    # Take spectral quantity
    s_result = s_result.take(fit_var)

    # Apply offset stated in "model"
    if "offset" in model:
        off_val, off_unit = model["offset"].split()
        s_result = s_result.offset(float(off_val), off_unit)

    # Apply normalization
    if "normalize" in pipeline:
        if pipeline["normalize"]:
            s_result = s_result.normalize()

    # Log the time mark when fitting is over
    end_fitting = time.time()
    time_fitting = float(end_fitting - begin_fitting)

    if verbose:
        print(f"\nSuccessfully finished the fitting process in {time_fitting}s.")

    log["time_fitting"] = time_fitting

    # PLOT THE DIFFERENCE BETWEEN THE TWO

    # If show_plot = True by default, show the diff plot
    if show_plot:
        plot_diff(
            s_data,
            s_result,
            fit_var,
            method=["diff", "ratio"],
            show=True,
        )

    if verbose:
        print(
            "\n======================== END OF FITTING PROCESS ========================\n"
        )

    return s_result, result, log



#cython: language_level=3
import numpy as np
cimport numpy as np

import cython
from cpython cimport array
from cython.operator import dereference

import ctypes
from libcpp cimport bool
from libcpp.map cimport map as mapcpp
from libcpp.set cimport set
from libcpp.utility cimport pair
from libcpp.vector cimport vector
from libc.string cimport memcpy
from libc.math cimport log
#from libcpp.functional cimport greater

# std::greater is currently not implemented in cython (but will be in the future)
# for now we have to import it in this way:
#cdef extern from "<functional>" namespace "std" nogil:
#    # Comparisons
#    cdef cppclass greater[T=*]:
#        greater() except +
#        bool operator()(const T& lhs, const T& rhs) except +

cdef extern from *:
    """
    struct greatercpp {
        bool operator () (const float x, const float y) const {return x > y;}
    };
    """
    ctypedef struct greatercpp:
        float a
        float b


#This can be modified to see if the compiled/built/installed version is current:
__version__ = 0.021

# Fast compiled version of add_at():
def add_at(arr, k,l,m, I):
    if arr.dtype != I.dtype:
        I = I.astype(arr.dtype)
    if arr.dtype == np.float32:
        return add_at_32(arr, k, l, m, I)
    else:
        return add_at_64(arr, k, l, m, I)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
def add_at_64(np.ndarray[np.float64_t, ndim=3] arr,
              np.ndarray[np.int32_t, ndim=1] k_arr,
              np.ndarray[np.int32_t, ndim=1] l_arr,
              np.ndarray[np.int32_t, ndim=1] m_arr,
              np.ndarray[np.float64_t, ndim=1] values):

    cdef unsigned int N = values.size
    cdef unsigned int i

    for i in range(N):
        arr[k_arr[i],l_arr[i],m_arr[i]] += values[i]

    return arr


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
def add_at_32(np.ndarray[np.float32_t, ndim=3] arr,
              np.ndarray[np.int32_t, ndim=1] k_arr,
              np.ndarray[np.int32_t, ndim=1] l_arr,
              np.ndarray[np.int32_t, ndim=1] m_arr,
              np.ndarray[np.float32_t, ndim=1] values):

    cdef unsigned int N = values.size
    cdef unsigned int i

    for i in range(N):
        arr[k_arr[i],l_arr[i],m_arr[i]] += values[i]

    return arr



# Fast compiled functions for determining the min/max broadening envelope.


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
def calc_lorentzian_envelope_params(
    np.ndarray[dtype=np.float32_t, ndim=1] na,
    np.ndarray[dtype=np.float32_t, ndim=1] gamma,
    verbose_gpu = False):

    cdef set[pair[float,float]] unique_set
    cdef float float_pair[2]

    cdef vector[pair[float,float]] duplicates_removed
    cdef vector[float] na_short
    cdef vector[float] gamma_short
    cdef mapcpp[float, float, greatercpp] bottom_envelope_map
    cdef mapcpp[float, float] top_envelope_map


    cdef float FLOAT_MAX = <float> 1e30
    cdef float FLOAT_MIN = <float>-1e30

    if verbose_gpu >= 2:
        print("Initializing Lorentzian parameters ")

    cdef size_t top_size = 0
    cdef size_t bottom_size = 0

    fname = "Lorenzian_minmax_" + str(len(gamma)) + ".dat"

    cdef unsigned int i,na_len
    cdef float na_i, gamma_i

    na_len = na.size
    for i in range(na_len):
        # Somewhat of a hack; all of the structs I tried show Python interaction
        # when storing values. float[2] didn't have this problem, so I fill the
        # float[2] array, and then convert the float* pointer to a const
        # pair[float,float]* pointer, which is then dereferenced by the [0].
        # well it gets the job done I suppose, there are no yellow lines inside
        # of this loop anymore.

        float_pair[0] = na[i]
        float_pair[1] = gamma[i]
        unique_set.insert((<const pair[float,float]*>float_pair)[0])

    duplicates_removed.assign(unique_set.begin(), unique_set.end())

    # make two new vectors where all duplicates are removed:
    for na_i, gamma_i in duplicates_removed:
        na_short.push_back(na_i)
        gamma_short.push_back(gamma_i)

    # identify candidates that might be part of the envelope:
    for i in range(len(na_short)):
        na_i = na_short[i]
        gamma_i = gamma_short[i]

        if bottom_envelope_map.count(na_i):
            if gamma_i < bottom_envelope_map.at(na_i):
                bottom_envelope_map[na_i] = gamma_i
        else:
            bottom_envelope_map.insert({na_i, gamma_i})

        if top_envelope_map.count(na_i):
            if gamma_i > top_envelope_map.at(na_i):
                top_envelope_map[na_i] = gamma_i
        else:
            top_envelope_map.insert({na_i, gamma_i})

    # For all candidates check which ones are actually part of the envelope:
    # First for the top:
    top_a = [dereference(top_envelope_map.begin()).first]
    top_b = [log(dereference(top_envelope_map.begin()).second)]
    top_x = [FLOAT_MIN]

    idx = 0
    for first_el, second_el in top_envelope_map:
        if idx != 0:
            for i in range(len(top_x)):
                x_ij = (log(second_el) - top_b[i]) / (top_a[i] - first_el)
                if x_ij >= top_x[i]:
                    if i < len(top_x) - 1:
                        if x_ij < top_x[i+1]:
                            break;
                    else:
                        break

            top_a = top_a[:i+1] + [first_el]
            top_b = top_b[:i+1] + [log(second_el)]
            top_x = top_x[:i+1] + [x_ij]

        idx+=1

    top_x = top_x[1:] + [FLOAT_MAX]

    #Then for the bottom:
    bottom_a = [dereference(bottom_envelope_map.begin()).first]
    bottom_b = [log(dereference(bottom_envelope_map.begin()).second)]
    bottom_x = [FLOAT_MIN]

    idx = 0
    for first_el, second_el in bottom_envelope_map:
        if idx != 0:
            for i in range(len(bottom_x)):
                x_ij = (log(second_el) - bottom_b[i]) / (bottom_a[i] - first_el)
                if x_ij >= bottom_x[i]:
                    if i < len(bottom_x) - 1:
                        if x_ij < bottom_x[i+1]:
                            break
                    else:
                        break

            bottom_a = bottom_a[:i+1] + [first_el]
            bottom_b = bottom_b[:i+1] + [log(second_el)]
            bottom_x = bottom_x[:i+1] + [x_ij]

        idx+=1

    bottom_x = bottom_x[1:] + [FLOAT_MAX]

    return ((bottom_a, bottom_b, bottom_x),
            (top_a, top_b, top_x))


def calc_gaussian_envelope_params(
    np.ndarray[dtype=np.float32_t, ndim=1] log_2vMm,
    verbose_gpu=False):

    log_2vMm_min = np.amin(log_2vMm)
    log_2vMm_max = np.amax(log_2vMm)

    return log_2vMm_min, log_2vMm_max

# Below are CPU functions extracted from the GPU module

cimport cpu_gpu_agnostic as cga
#from radis.lbl.gpu import initData, iterData

cpdef void fillLDM(gridDim, blockDim, args):

   cdef np.ndarray[np.uint8_t, ndim=1] iso
   cdef np.ndarray[np.float32_t, ndim=1] v0
   cdef np.ndarray[np.float32_t, ndim=1] da
   cdef np.ndarray[np.float32_t, ndim=1] S0
   cdef np.ndarray[np.float32_t, ndim=1] El
   cdef np.ndarray[np.float32_t, ndim=1] gamma
   cdef np.ndarray[np.float32_t, ndim=1] na
   cdef np.ndarray[np.float32_t, ndim=3] S_klm

   iso, v0, da, S0, El, gamma, na, S_klm = args

   cga.set_dims(blockDim[0],gridDim[0])
   cga.fillLDM(&iso[0], &v0[0], &da[0], &S0[0], &El[0], &gamma[0], &na[0], &S_klm[0,0,0])


cpdef void applyLineshapes(gridDim, blockDim, args):

    cdef np.ndarray[np.complex64_t, ndim=3] S_klm_FT
    cdef np.ndarray[np.complex64_t, ndim=1] abscoeff
    S_klm_FT, abscoeff = args

    cga.set_dims(blockDim[0], gridDim[0])
    cga.applyLineshapes(&S_klm_FT[0,0,0], &abscoeff[0])


cpdef void calcTransmittanceNoslit(gridDim, blockDim, args):

    cdef np.ndarray[np.float32_t, ndim=1] abscoeff,
    cdef np.ndarray[np.float32_t, ndim=1] transmittance_noslit
    abscoeff, transmittance_noslit = args

    cga.set_dims(blockDim[0], gridDim[0])
    cga.calcTransmittanceNoslit(&abscoeff[0], &transmittance_noslit[0])


cpdef void applyGaussianSlit(gridDim, blockDim, args):

    cdef np.ndarray[np.complex64_t, ndim=1] transmittance_noslit_FT,
    cdef np.ndarray[np.complex64_t, ndim=1] transmittance_FT
    transmittance_noslit_FT, transmittance_FT = args

    cga.set_dims(blockDim[0],gridDim[0])
    cga.applyGaussianSlit(&transmittance_noslit_FT[0], &transmittance_FT[0])


cpdef set_init_params(params):
    cdef void* cpp_ptr = cga.get_init_ptr()
    cdef size_t size = ctypes.sizeof(params)
    cdef char[:] struct_memview = bytearray(params)
    memcpy(cpp_ptr, &struct_memview[0], size)


cpdef set_iter_params(params):
    cdef void* cpp_ptr = cga.get_iter_ptr()
    cdef size_t size = ctypes.sizeof(params)
    cdef char[:] struct_memview = bytearray(params)
    memcpy(cpp_ptr, &struct_memview[0], size)

cpdef get_T():
    return cga.get_T()



from numpy cimport complex64_t

cdef extern from "../lbl/gpu.h":

    void fillLDM(unsigned char* iso, float* v0, float* da, float* S0, float* El, float* gamma, float* na, float* S_klm);
    void applyLineshapes(complex64_t* S_klm_FT, complex64_t* abscoeff);
    void calcTransmittanceNoslit(float* abscoeff, float* transmittance_noslit);
    void applyGaussianSlit(complex64_t* transmittance_noslit_FT, complex64_t* transmittance_FT);
    void* get_init_ptr();
    void* get_iter_ptr();
    void set_dims(int blockDim_x, int gridDim_x);
    float get_T();
